@@ -364,8 +364,8 @@ static gpg_error_t
 load_certs_from_dir (const char *dirname, unsigned int trustclass)
 {
   gpg_error_t err;
-  DIR *dir;
-  struct dirent *ep;
+  gnupg_dir_t dir;
+  gnupg_dirent_t ep;
   char *p;
   size_t n;
   estream_t fp;
@@ -373,13 +373,13 @@ load_certs_from_dir (const char *dirname, unsigned int trustclass)
   ksba_cert_t cert;
   char *fname = NULL;
 
-  dir = opendir (dirname);
+  dir = gnupg_opendir (dirname);
   if (!dir)
     {
       return 0; /* We do not consider this a severe error.  */
     }
 
-  while ( (ep=readdir (dir)) )
+  while ( (ep = gnupg_readdir (dir)) )
     {
       p = ep->d_name;
       if (*p == '.' || !*p)
@@ -423,6 +423,9 @@ load_certs_from_dir (const char *dirname, unsigned int trustclass)
         log_info (_("certificate '%s' already cached\n"), fname);
       else if (!err)
         {
+          if ((trustclass & CERTTRUST_CLASS_CONFIG))
+            http_register_cfg_ca (fname);
+
           if (trustclass)
             log_info (_("trusted certificate '%s' loaded\n"), fname);
           else
@@ -444,7 +447,7 @@ load_certs_from_dir (const char *dirname, unsigned int trustclass)
     }
 
   xfree (fname);
-  closedir (dir);
+  gnupg_closedir (dir);
   return 0;
 }
 
@@ -569,11 +572,11 @@ load_certs_from_w32_store (const char *storename)
         }
 
       pCertOpenSystemStore = (CERTOPENSYSTEMSTORE)
-        GetProcAddress (hCrypt32, "CertOpenSystemStoreA");
+        (void*)GetProcAddress (hCrypt32, "CertOpenSystemStoreA");
       pCertEnumCertificatesInStore = (CERTENUMCERTIFICATESINSTORE)
-        GetProcAddress (hCrypt32, "CertEnumCertificatesInStore");
+        (void*)GetProcAddress (hCrypt32, "CertEnumCertificatesInStore");
       pCertCloseStore = (CERTCLOSESTORE)
-        GetProcAddress (hCrypt32, "CertCloseStore");
+        (void*)GetProcAddress (hCrypt32, "CertCloseStore");
       if (   !pCertOpenSystemStore
           || !pCertEnumCertificatesInStore
           || !pCertCloseStore)
@@ -682,7 +685,7 @@ load_certs_from_system (void)
   gpg_error_t err = 0;
 
   for (idx=0; idx < DIM (table); idx++)
-    if (!access (table[idx].name, F_OK))
+    if (!gnupg_access (table[idx].name, F_OK))
       {
         /* Take the first available bundle.  */
         err = load_certs_from_file (table[idx].name, CERTTRUST_CLASS_SYSTEM, 0);
@@ -718,11 +721,15 @@ cert_cache_init (strlist_t hkp_cacerts)
     load_certs_from_dir (fname, 0);
   xfree (fname);
 
-  fname = make_filename_try (gnupg_datadir (),
-                             "sks-keyservers.netCA.pem", NULL);
-  if (fname)
-    load_certs_from_file (fname, CERTTRUST_CLASS_HKPSPOOL, 1);
-  xfree (fname);
+  /* Put the special pool certificate into our store.  This is
+   * currently only used with ntbtls.  For GnuTLS http_session_new
+   * unfortunately loads that certificate directly from the file.  */
+  /* Disabled for 2.3.2 because the service had to be shutdown.  */
+  /* fname = make_filename_try (gnupg_datadir (), */
+  /*                            "sks-keyservers.netCA.pem", NULL); */
+  /* if (fname) */
+  /*   load_certs_from_file (fname, CERTTRUST_CLASS_HKPSPOOL, 1); */
+  /* xfree (fname); */
 
   for (sl = hkp_cacerts; sl; sl = sl->next)
     load_certs_from_file (sl->d, CERTTRUST_CLASS_HKP, 0);
@@ -763,6 +770,8 @@ cert_cache_deinit (int full)
         }
     }
 
+  http_register_cfg_ca (NULL);
+
   total_nonperm_certificates = 0;
   any_cert_of_class = 0;
   initialization_done = 0;
@@ -1466,6 +1475,9 @@ find_cert_bysubject (ctrl_t ctrl, const char *subject_dn, ksba_sexp_t keyid)
                 {
                   ksba_cert_ref (ci->cert);
                   release_cache_lock ();
+                  if (DBG_LOOKUP)
+                    log_debug ("%s: certificate found in the cache"
+                               " via ocsp_certs\n", __func__);
                   return ci->cert; /* We use this certificate. */
                 }
       release_cache_lock ();
@@ -1473,7 +1485,7 @@ find_cert_bysubject (ctrl_t ctrl, const char *subject_dn, ksba_sexp_t keyid)
         log_debug ("find_cert_bysubject: certificate not in ocsp_certs\n");
     }
 
-  /* No check whether the certificate is cached.  */
+  /* Now check whether the certificate is cached.  */
   for (seq=0; (cert = get_cert_bysubject (subject_dn, seq)); seq++)
     {
       if (!keyid)
@@ -1482,6 +1494,9 @@ find_cert_bysubject (ctrl_t ctrl, const char *subject_dn, ksba_sexp_t keyid)
           && !cmp_simple_canon_sexp (keyid, subj))
         {
           xfree (subj);
+          if (DBG_LOOKUP)
+            log_debug ("%s: certificate found in the cache"
+                       " via subject DN\n", __func__);
           break; /* Found matching cert. */
         }
       xfree (subj);
@@ -1490,6 +1505,34 @@ find_cert_bysubject (ctrl_t ctrl, const char *subject_dn, ksba_sexp_t keyid)
   if (cert)
     return cert; /* Done.  */
 
+  /* If we do not have a subject DN but have a keyid, try to locate it
+   * by keyid.  */
+  if (!subject_dn && keyid)
+    {
+      int i;
+      cert_item_t ci;
+      ksba_sexp_t ski;
+
+      acquire_cache_read_lock ();
+      for (i=0; i < 256; i++)
+        for (ci=cert_cache[i]; ci; ci = ci->next)
+          if (ci->cert && !ksba_cert_get_subj_key_id (ci->cert, NULL, &ski))
+            {
+              if (!cmp_simple_canon_sexp (keyid, ski))
+                {
+                  ksba_free (ski);
+                  ksba_cert_ref (ci->cert);
+                  release_cache_lock ();
+                  if (DBG_LOOKUP)
+                    log_debug ("%s: certificate found in the cache"
+                               " via ski\n", __func__);
+                  return ci->cert;
+                }
+              ksba_free (ski);
+            }
+      release_cache_lock ();
+    }
+
   if (DBG_LOOKUP)
     log_debug ("find_cert_bysubject: certificate not in cache\n");
 
