@@ -1625,21 +1625,8 @@ start_sig_check (ksba_crl_t crl, gcry_md_hd_t *md, int *algo, int *use_pss)
     }
   else
     *algo = gcry_md_map_name (algoid);
-  if (!*algo && algoid)
-    {
-      if (!strcmp (algoid, "1.2.840.10045.4.3.1"))
-        *algo = GCRY_MD_SHA224; /* ecdsa-with-sha224 */
-      else if (!strcmp (algoid, "1.2.840.10045.4.3.2"))
-        *algo = GCRY_MD_SHA256; /* ecdsa-with-sha256 */
-      else if (!strcmp (algoid, "1.2.840.10045.4.3.3"))
-        *algo = GCRY_MD_SHA384; /* ecdsa-with-sha384 */
-      else if (!strcmp (algoid, "1.2.840.10045.4.3.4"))
-        *algo = GCRY_MD_SHA512; /* ecdsa-with-sha512 */
-    }
   if (!*algo)
     {
-      log_debug ("XXXXX %s: %s <%s>\n",
-                 __func__, gpg_strerror (err), gpg_strsource (err));
       log_error (_("unknown hash algorithm '%s'\n"), algoid? algoid:"?");
       return gpg_error (GPG_ERR_DIGEST_ALGO);
     }
@@ -1673,7 +1660,6 @@ finish_sig_check (ksba_crl_t crl, gcry_md_hd_t md, int algo,
   size_t n;
   gcry_sexp_t s_sig = NULL, s_hash = NULL, s_pkey = NULL;
   unsigned int saltlen = 0;  /* (used only with use_pss)  */
-  int pkalgo;
 
   /* This also stops debugging on the MD.  */
   gcry_md_final (md);
@@ -1739,7 +1725,8 @@ finish_sig_check (ksba_crl_t crl, gcry_md_hd_t md, int algo,
         {
           log_error ("hash algo mismatch: %d announced but %d used\n",
                      algo, hashalgo);
-          return gpg_error (GPG_ERR_INV_CRL);
+          err = gpg_error (GPG_ERR_INV_CRL);
+          goto leave;
         }
       /* Add some restrictions; see ../sm/certcheck.c for details.  */
       switch (algo)
@@ -1755,14 +1742,16 @@ finish_sig_check (ksba_crl_t crl, gcry_md_hd_t md, int algo,
         default:
           log_error ("PSS hash algorithm '%s' rejected\n",
                      gcry_md_algo_name (algo));
-          return gpg_error (GPG_ERR_DIGEST_ALGO);
+          err = gpg_error (GPG_ERR_DIGEST_ALGO);
+          goto leave;
         }
 
       if (gcry_md_get_algo_dlen (algo) != saltlen)
         {
           log_error ("PSS hash algorithm '%s' rejected due to salt length %u\n",
                      gcry_md_algo_name (algo), saltlen);
-          return gpg_error (GPG_ERR_DIGEST_ALGO);
+          err = gpg_error (GPG_ERR_DIGEST_ALGO);
+          goto leave;
         }
     }
 
@@ -1798,54 +1787,6 @@ finish_sig_check (ksba_crl_t crl, gcry_md_hd_t md, int algo,
                              gcry_md_read (md, algo),
                              saltlen);
     }
-  else if ((pkalgo = pk_algo_from_sexp (s_pkey)) == GCRY_PK_ECC)
-    {
-      unsigned int qbits0, qbits;
-
-      qbits0 = gcry_pk_get_nbits (s_pkey);
-      qbits = qbits0 == 521? 512 : qbits0;
-
-      if ((qbits%8))
-	{
-	  log_error ("ECDSA requires the hash length to be a"
-                     " multiple of 8 bits\n");
-	  err = gpg_error (GPG_ERR_INTERNAL);
-          goto leave;
-	}
-
-      /* Don't allow any Q smaller than 160 bits.  */
-      if (qbits < 160)
-	{
-	  log_error (_("%s key uses an unsafe (%u bit) hash\n"),
-                     gcry_pk_algo_name (pkalgo), qbits0);
-	  err = gpg_error (GPG_ERR_INTERNAL);
-          goto leave;
-	}
-
-      /* Check if we're too short.  */
-      n = gcry_md_get_algo_dlen (algo);
-      if (n < qbits/8)
-        {
-	  log_error (_("a %u bit hash is not valid for a %u bit %s key\n"),
-                     (unsigned int)n*8,
-                     qbits0,
-                     gcry_pk_algo_name (pkalgo));
-          if (n < 20)
-            {
-              err = gpg_error (GPG_ERR_INTERNAL);
-              goto leave;
-            }
-        }
-
-      /* Truncate.  */
-      if (n > qbits/8)
-        n = qbits/8;
-
-      err = gcry_sexp_build (&s_hash, NULL, "(data(flags raw)(value %b))",
-                             (int)n,
-                             gcry_md_read (md, algo));
-
-    }
   else
     {
       err = gcry_sexp_build (&s_hash, NULL,
@@ -1863,7 +1804,7 @@ finish_sig_check (ksba_crl_t crl, gcry_md_hd_t md, int algo,
   /* Pass this on to the signature verification. */
   err = gcry_pk_verify (s_sig, s_hash, s_pkey);
   if (DBG_X509)
-    log_debug ("%s: gcry_pk_verify: %s\n", __func__, gpg_strerror (err));
+    log_debug ("gcry_pk_verify: %s\n", gpg_strerror (err));
 
  leave:
   xfree (sigval);
@@ -2012,7 +1953,7 @@ crl_parse_insert (ctrl_t ctrl, ksba_crl_t crl,
             ksba_sexp_t keyid;
 
             /* We need to look for the issuer only after having read
-               all items.  The issuer itselfs comes before the items
+               all items.  The issuer itself comes before the items
                but the optional authorityKeyIdentifier comes after the
                items. */
             err = ksba_crl_get_issuer (crl, &crlissuer);
@@ -2137,7 +2078,7 @@ get_crl_number (ksba_crl_t crl)
 
 
 /* Return the authorityKeyIdentifier or NULL if it is not available.
-   The issuer name may consists of several parts - they are delimted by
+   The issuer name may consists of several parts - they are delimited by
    0x01. */
 static char *
 get_auth_key_id (ksba_crl_t crl, char **serialno)
@@ -2356,21 +2297,11 @@ crl_cache_insert (ctrl_t ctrl, const char *url, ksba_reader_t reader)
   for (idx=0; !(err=ksba_crl_get_extension (crl, idx, &oid, &critical,
                                               NULL, NULL)); idx++)
     {
-      strlist_t sl;
-
       if (!critical
           || !strcmp (oid, oidstr_authorityKeyIdentifier)
           || !strcmp (oid, oidstr_crlNumber) )
         continue;
-
-      for (sl=opt.ignored_crl_extensions;
-           sl && strcmp (sl->d, oid); sl = sl->next)
-        ;
-      if (sl)
-        continue;  /* Is in ignored list.  */
-
       log_error (_("unknown critical CRL extension %s\n"), oid);
-      log_info ("(CRL='%s')\n", url);
       if (!err2)
         err2 = gpg_error (GPG_ERR_INV_CRL);
       invalidate_crl |= 2;
@@ -2683,21 +2614,21 @@ crl_cache_reload_crl (ctrl_t ctrl, ksba_cert_t cert)
   ksba_name_t distpoint = NULL;
   ksba_name_t issuername = NULL;
   char *distpoint_uri = NULL;
+  char *issuername_uri = NULL;
   int any_dist_point = 0;
   int seq;
-  gpg_error_t last_err = 0;
 
   /* Loop over all distribution points, get the CRLs and put them into
      the cache. */
   if (opt.verbose)
     log_info ("checking distribution points\n");
   seq = 0;
-  while (xfree (distpoint), xfree (issuername),
-         !(err = ksba_cert_get_crl_dist_point (cert, seq++,
+  while ( !(err = ksba_cert_get_crl_dist_point (cert, seq++,
                                                 &distpoint,
                                                 &issuername, NULL )))
     {
       int name_seq;
+      gpg_error_t last_err = 0;
 
       if (!distpoint && !issuername)
         {
@@ -2708,11 +2639,13 @@ crl_cache_reload_crl (ctrl_t ctrl, ksba_cert_t cert)
                     suitable CRL. */
         }
 
+      xfree (issuername_uri); issuername_uri = NULL;
+
       /* Get the URIs.  We do this in a loop to iterate over all names
          in the crlDP. */
       for (name_seq=0; ksba_name_enum (distpoint, name_seq); name_seq++)
         {
-          xfree (distpoint_uri);
+          xfree (distpoint_uri); distpoint_uri = NULL;
           distpoint_uri = ksba_name_get_uri (distpoint, name_seq);
           if (!distpoint_uri)
             continue;
@@ -2736,7 +2669,6 @@ crl_cache_reload_crl (ctrl_t ctrl, ksba_cert_t cert)
 
           if (opt.verbose)
             log_info ("fetching CRL from '%s'\n", distpoint_uri);
-          crl_close_reader (reader);
           err = crl_fetch (ctrl, distpoint_uri, &reader);
           if (err)
             {
@@ -2756,16 +2688,28 @@ crl_cache_reload_crl (ctrl_t ctrl, ksba_cert_t cert)
               last_err = err;
               continue; /* with the next name. */
             }
-          goto leave; /* Ready - we got the CRL. */
+          last_err = 0;
+          break; /* Ready. */
         }
+      if (last_err)
+        {
+          err = last_err;
+          goto leave;
+        }
+
+      ksba_name_release (distpoint); distpoint = NULL;
+
+      /* We don't do anything with issuername_uri yet but we keep the
+         code for documentation. */
+      issuername_uri =  ksba_name_get_uri (issuername, 0);
+      ksba_name_release (issuername); issuername = NULL;
+
+      /* Close the reader.  */
+      crl_close_reader (reader);
+      reader = NULL;
     }
   if (gpg_err_code (err) == GPG_ERR_EOF)
     err = 0;
-  if (!err && last_err)
-    {
-      err = last_err;
-      goto leave;
-    }
 
   /* If we did not found any distpoint, try something reasonable. */
   if (!any_dist_point )
@@ -2773,6 +2717,9 @@ crl_cache_reload_crl (ctrl_t ctrl, ksba_cert_t cert)
       if (opt.verbose)
         log_info ("no distribution point - trying issuer name\n");
 
+      crl_close_reader (reader);
+      reader = NULL;
+
       issuer = ksba_cert_get_issuer (cert, 0);
       if (!issuer)
         {
@@ -2783,7 +2730,6 @@ crl_cache_reload_crl (ctrl_t ctrl, ksba_cert_t cert)
 
       if (opt.verbose)
         log_info ("fetching CRL from default location\n");
-      crl_close_reader (reader);
       err = crl_fetch_default (ctrl, issuer, &reader);
       if (err)
           {
@@ -2806,6 +2752,7 @@ crl_cache_reload_crl (ctrl_t ctrl, ksba_cert_t cert)
  leave:
   crl_close_reader (reader);
   xfree (distpoint_uri);
+  xfree (issuername_uri);
   ksba_name_release (distpoint);
   ksba_name_release (issuername);
   ksba_free (issuer);
