@@ -71,6 +71,7 @@
 #if USE_LDAP
 # include "ldap-wrapper.h"
 #endif
+#include "../common/comopt.h"
 #include "../common/init.h"
 #include "../common/gc-opt-flags.h"
 #include "dns-stuff.h"
@@ -124,6 +125,7 @@ enum cmd_and_opt_values {
   oHTTPProxy,
   oLDAPProxy,
   oOnlyLDAPProxy,
+  oLDAPServer,
   oLDAPFile,
   oLDAPTimeout,
   oLDAPAddServers,
@@ -259,6 +261,8 @@ static gpgrt_opt_t opts[] = {
                 N_("|HOST|use HOST for LDAP queries")),
   ARGPARSE_s_n (oOnlyLDAPProxy, "only-ldap-proxy",
                 N_("do not use fallback hosts with --ldap-proxy")),
+  ARGPARSE_s_s (oLDAPServer, "ldapserver",
+                N_("|SPEC|use this keyserver to lookup keys")),
   ARGPARSE_s_s (oLDAPFile, "ldapserverlist-file",
                 N_("|FILE|read LDAP server list from FILE")),
   ARGPARSE_s_n (oLDAPAddServers, "add-servers",
@@ -377,6 +381,11 @@ static int network_activity_seen;
 /* A list of filenames registered with --hkp-cacert.  */
 static strlist_t hkp_cacert_filenames;
 
+/* A flag used to clear the list of ldapservers iff --ldapserver is
+ * given on the command line or one of the conf files. In this case we
+ * want to clear all old specifications through the legacy
+ * dirmngr_ldapservers.conf. */
+static int ldapserver_list_needs_reset;
 
 /* The timer tick used for housekeeping stuff.  The second constant is used when a shutdown is pending.  */
 #define TIMERTICK_INTERVAL           (60)
@@ -670,6 +679,8 @@ parse_rereadable_options (gpgrt_argparse_t *pargs, int reread)
       set_dns_timeout (0);
       opt.connect_timeout = 0;
       opt.connect_quick_timeout = 0;
+      opt.ldaptimeout = DEFAULT_LDAP_TIMEOUT;
+      ldapserver_list_needs_reset = 1;
       return 1;
     }
 
@@ -758,6 +769,32 @@ parse_rereadable_options (gpgrt_argparse_t *pargs, int reread)
     case oStandardResolver: enable_standard_resolver (1); break;
     case oRecursiveResolver: enable_recursive_resolver (1); break;
 
+    case oLDAPServer:
+#if USE_LDAP
+      {
+        ldap_server_t server;
+        char *p;
+
+        p = pargs->r.ret_str;
+        if (!strncmp (p, "ldap:", 5) && !(p[5] == '/' && p[6] == '/'))
+          p += 5;
+
+        server = ldapserver_parse_one (p, NULL, 0);
+        if (server)
+          {
+            if (ldapserver_list_needs_reset)
+              {
+                ldapserver_list_needs_reset = 0;
+                ldapserver_list_free (opt.ldapservers);
+                opt.ldapservers = NULL;
+              }
+            server->next = opt.ldapservers;
+            opt.ldapservers = server;
+          }
+      }
+#endif
+      break;
+
     case oKeyServer:
       if (*pargs->r.ret_str)
         add_to_strlist (&opt.keyserver, pargs->r.ret_str);
@@ -779,6 +816,10 @@ parse_rereadable_options (gpgrt_argparse_t *pargs, int reread)
       opt.connect_quick_timeout = pargs->r.ret_ulong * 1000;
       break;
 
+    case oLDAPTimeout:
+      opt.ldaptimeout = pargs->r.ret_int;
+      break;
+
     default:
       return 0; /* Not handled. */
     }
@@ -1034,9 +1075,6 @@ main (int argc, char **argv)
 #        endif /*USE_LDAP*/
           break;
 	case oLDAPAddServers: opt.add_new_ldapservers = 1; break;
-	case oLDAPTimeout:
-	  opt.ldaptimeout = pargs.r.ret_int;
-	  break;
 
         case oFakedSystemTime:
           gnupg_set_time ((time_t)pargs.r.ret_ulong, 0);
@@ -1072,6 +1110,14 @@ main (int argc, char **argv)
 
   if (log_get_errorcount(0))
     exit(2);
+
+  /* Get a default log file from common.conf.  */
+  if (!logfile && !parse_comopt (GNUPG_MODULE_NAME_DIRMNGR, debug_argparser))
+    {
+      logfile = comopt.logfile;
+      comopt.logfile = NULL;
+    }
+
   if (nogreeting )
     greeting = 0;
 
@@ -1124,9 +1170,11 @@ main (int argc, char **argv)
   if (cmd != aGPGConfTest && cmd != aGPGConfList && cmd != aGPGConfVersions)
     set_tor_mode ();
 
-  /* Get LDAP server list from file. */
+  /* Get LDAP server list from file unless --ldapserver has been used.  */
 #if USE_LDAP
-  if (!ldapfile)
+  if (opt.ldapservers)
+    ;
+  else if (!ldapfile)
     {
       ldapfile = make_filename (gnupg_homedir (),
                                 "dirmngr_ldapservers.conf",
@@ -1806,9 +1854,10 @@ reread_configuration (void)
   gpgrt_argparse_t pargs;
   char *twopart;
   int dummy;
+  int logfile_seen = 0;
 
   if (!opt.config_filename)
-    return; /* No config file. */
+    goto finish; /* No config file. */
 
   twopart = strconcat (DIRMNGR_NAME EXTSEP_S "conf" PATHSEP_S,
                        opt.config_filename, NULL);
@@ -1833,11 +1882,28 @@ reread_configuration (void)
       else if (pargs.r_opt < -1)
         pargs.err = ARGPARSE_PRINT_WARNING;
       else /* Try to parse this option - ignore unchangeable ones. */
-        parse_rereadable_options (&pargs, 1);
+        {
+          if (pargs.r_opt == oLogFile)
+            logfile_seen = 1;
+          parse_rereadable_options (&pargs, 1);
+        }
     }
   gpgrt_argparse (NULL, &pargs, NULL);  /* Release internal state.  */
   xfree (twopart);
   post_option_parsing ();
+
+ finish:
+  /* Get a default log file from common.conf.  */
+  if (!logfile_seen && !parse_comopt (GNUPG_MODULE_NAME_DIRMNGR, !!opt.debug))
+    {
+      if (!current_logfile || !comopt.logfile
+          || strcmp (current_logfile, comopt.logfile))
+        {
+          log_set_file (comopt.logfile);
+          xfree (current_logfile);
+          current_logfile = comopt.logfile? xtrystrdup (comopt.logfile) : NULL;
+        }
+    }
 }
 
 
