@@ -50,7 +50,6 @@ struct ldap_server_s
   char *user;
   char *pass;
   char *base;
-
   unsigned int starttls:1;       /* Use STARTTLS.  */
   unsigned int ldap_over_tls:1;  /* Use LDAP over an TLS tunnel */
   unsigned int ntds:1;           /* Use Active Directory authentication.  */
@@ -223,7 +222,7 @@ int dirmngr_use_tor (void);
 /*-- Various housekeeping functions.  --*/
 void ks_hkp_housekeeping (time_t curtime);
 void ks_hkp_reload (void);
-
+void ks_hkp_init (void);
 
 /*-- server.c --*/
 ldap_server_t get_ldapservers_from_ctrl (ctrl_t ctrl);
@@ -237,8 +236,6 @@ int dirmngr_assuan_log_monitor (assuan_context_t ctx, unsigned int cat,
 void start_command_handler (gnupg_fd_t fd, unsigned int session_id);
 gpg_error_t dirmngr_tick (ctrl_t ctrl);
 
-/* (See also dirmngr-status.h)  */
-
 /*-- http-ntbtls.c --*/
 /* Note that we don't use a callback for gnutls.  */
 
