@@ -1,7 +1,7 @@
 /* gpgsm.c - GnuPG for S/MIME
  * Copyright (C) 2001-2020 Free Software Foundation, Inc.
  * Copyright (C) 2001-2019 Werner Koch
- * Copyright (C) 2015-2020 g10 Code GmbH
+ * Copyright (C) 2015-2021 g10 Code GmbH
  *
  * This file is part of GnuPG.
  *
@@ -28,6 +28,8 @@
 #include <ctype.h>
 #include <unistd.h>
 #include <fcntl.h>
+/*#include <mcheck.h>*/
+#include <npth.h>
 
 #define INCLUDED_BY_MAIN_MODULE 1
 
@@ -45,9 +47,9 @@
 #include "../common/asshelp.h"
 #include "../common/init.h"
 #include "../common/compliance.h"
+#include "../common/comopt.h"
 #include "minip12.h"
 
-
 #ifndef O_BINARY
 #define O_BINARY 0
 #endif
@@ -94,6 +96,7 @@ enum cmd_and_opt_values {
   aDumpChain,
   aDumpSecretKeys,
   aDumpExternalKeys,
+  aShowCerts,
   aKeydbClearSomeCertFlags,
   aFingerprint,
 
@@ -106,6 +109,7 @@ enum cmd_and_opt_values {
   oDebugAllowCoreDump,
   oDebugNoChainValidation,
   oDebugIgnoreExpiration,
+  oDebugForceECDHSHA1KDF,
   oLogFile,
   oNoLogFile,
   oAuditLog,
@@ -134,7 +138,6 @@ enum cmd_and_opt_values {
   oAssumeArmor,
   oAssumeBase64,
   oAssumeBinary,
-  oInputSizeHint,
 
   oBase64,
   oNoArmor,
@@ -157,12 +160,12 @@ enum cmd_and_opt_values {
   oDisablePolicyChecks,
   oEnablePolicyChecks,
   oAutoIssuerKeyRetrieve,
-  oMinRSALength,
 
   oWithFingerprint,
   oWithMD5Fingerprint,
   oWithKeygrip,
   oWithSecret,
+  oWithKeyScreening,
   oAnswerYes,
   oAnswerNo,
   oKeyring,
@@ -189,7 +192,6 @@ enum cmd_and_opt_values {
   oSkipVerify,
   oValidationModel,
   oKeyServer,
-  oKeyServer_deprecated,
   oEncryptTo,
   oNoEncryptTo,
   oLoggerFD,
@@ -199,16 +201,16 @@ enum cmd_and_opt_values {
   oNoRandomSeedFile,
   oNoCommonCertsImport,
   oIgnoreCertExtension,
-  oIgnoreCertWithOID,
-  oRequireCompliance,
-  oCompatibilityFlags,
-  oKbxBufferSize,
-  oAlwaysTrust,
+  oAuthenticode,
+  oAttribute,
+  oChUid,
+  oUseKeyboxd,
+  oKeyboxdProgram,
   oNoAutostart
  };
 
 
-static ARGPARSE_OPTS opts[] = {
+static gpgrt_opt_t opts[] = {
 
   ARGPARSE_group (300, N_("@Commands:\n ")),
 
@@ -251,6 +253,7 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_c (aGPGConfList, "gpgconf-list", "@"),
   ARGPARSE_c (aGPGConfTest, "gpgconf-test", "@"),
 
+  ARGPARSE_c (aShowCerts, "show-certs", "@"),
   ARGPARSE_c (aDumpKeys, "dump-cert", "@"),
   ARGPARSE_c (aDumpKeys, "dump-keys", "@"),
   ARGPARSE_c (aDumpChain, "dump-chain", "@"),
@@ -275,6 +278,7 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_s_n (oDebugAllowCoreDump, "debug-allow-core-dump", "@"),
   ARGPARSE_s_n (oDebugNoChainValidation, "debug-no-chain-validation", "@"),
   ARGPARSE_s_n (oDebugIgnoreExpiration,  "debug-ignore-expiration", "@"),
+  ARGPARSE_s_n (oDebugForceECDHSHA1KDF,  "debug-force-ecdh-sha1kdf", "@"),
   ARGPARSE_s_s (oLogFile, "log-file",
                 N_("|FILE|write server mode logs to FILE")),
   ARGPARSE_s_n (oNoLogFile, "no-log-file", "@"),
@@ -294,12 +298,11 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_s_s (oPolicyFile, "policy-file",
                 N_("|FILE|take policy information from FILE")),
   ARGPARSE_s_s (oCompliance, "compliance",   "@"),
-  ARGPARSE_p_u (oMinRSALength, "min-rsa-length", "@"),
   ARGPARSE_s_n (oNoCommonCertsImport, "no-common-certs-import", "@"),
   ARGPARSE_s_s (oIgnoreCertExtension, "ignore-cert-extension", "@"),
-  ARGPARSE_s_s (oIgnoreCertWithOID, "ignore-cert-with-oid", "@"),
   ARGPARSE_s_n (oNoAutostart, "no-autostart", "@"),
   ARGPARSE_s_s (oAgentProgram, "agent-program", "@"),
+  ARGPARSE_s_s (oKeyboxdProgram, "keyboxd-program", "@"),
   ARGPARSE_s_s (oDirmngrProgram, "dirmngr-program", "@"),
   ARGPARSE_s_s (oProtectToolProgram, "protect-tool-program", "@"),
 
@@ -312,7 +315,6 @@ static ARGPARSE_OPTS opts[] = {
                 N_("assume input is in base-64 format")),
   ARGPARSE_s_n (oAssumeBinary, "assume-binary",
                 N_("assume input is in binary format")),
-  ARGPARSE_s_s (oInputSizeHint, "input-size-hint", "@"),
 
 
   ARGPARSE_header ("Output", N_("Options controlling the output")),
@@ -323,6 +325,8 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_s_n (oNoArmor, "no-armour", "@"),
   ARGPARSE_s_n (oBase64, "base64", N_("create base-64 encoded output")),
   ARGPARSE_s_s (oOutput, "output", N_("|FILE|write output to FILE")),
+  ARGPARSE_s_n (oAuthenticode, "authenticode", "@"),
+  ARGPARSE_s_s (oAttribute,    "attribute", "@"),
 
 
   ARGPARSE_header (NULL, N_("Options to specify keys")),
@@ -344,8 +348,11 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_s_s (oKeyring, "keyring",
                 N_("|FILE|add keyring to the list of keyrings")),
   ARGPARSE_s_n (oNoDefKeyring, "no-default-keyring", "@"),
-  ARGPARSE_s_s (oKeyServer_deprecated, "ldapserver", "@"),
+  ARGPARSE_s_s (oKeyServer, "ldapserver",
+                N_("|SPEC|use this keyserver to lookup keys")),
   ARGPARSE_s_s (oKeyServer, "keyserver", "@"),
+  ARGPARSE_s_n (oUseKeyboxd,    "use-keyboxd", "@"),
+
 
   ARGPARSE_header ("ImportExport",
                    N_("Options controlling key import and export")),
@@ -369,6 +376,9 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_s_n (oWithFingerprint, "with-fingerprint", "@"),
   ARGPARSE_s_n (oWithKeygrip,     "with-keygrip", "@"),
   ARGPARSE_s_n (oWithSecret,      "with-secret", "@"),
+  ARGPARSE_s_n (oWithKeyScreening,"with-key-screening", "@"),
+
+
 
   ARGPARSE_header ("Security", N_("Options controlling the security")),
 
@@ -394,8 +404,6 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_s_s (oDisablePubkeyAlgo,  "disable-pubkey-algo", "@"),
   ARGPARSE_s_n (oIgnoreTimeConflict, "ignore-time-conflict", "@"),
   ARGPARSE_s_n (oNoRandomSeedFile,  "no-random-seed-file", "@"),
-  ARGPARSE_s_n (oRequireCompliance, "require-compliance", "@"),
-  ARGPARSE_s_n (oAlwaysTrust,       "always-trust", "@"),
 
 
   ARGPARSE_header (NULL, N_("Options for unattended use")),
@@ -428,8 +436,8 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_s_s (oLCctype,    "lc-ctype", "@"),
   ARGPARSE_s_s (oLCmessages, "lc-messages", "@"),
   ARGPARSE_s_s (oXauthority, "xauthority", "@"),
-  ARGPARSE_s_s (oCompatibilityFlags, "compatibility-flags", "@"),
-  ARGPARSE_p_u (oKbxBufferSize,  "kbx-buffer-size", "@"),
+  ARGPARSE_s_s (oChUid, "chuid", "@"),
+
 
   ARGPARSE_header (NULL, ""),  /* Stop the header group.  */
 
@@ -461,16 +469,8 @@ static struct debug_flags_s debug_flags [] =
     { DBG_MEMSTAT_VALUE, "memstat" },
     { DBG_HASHING_VALUE, "hashing" },
     { DBG_IPC_VALUE    , "ipc"     },
-    { 0, NULL }
-  };
-
-
-/* The list of compatibility flags.  */
-static struct compatibility_flags_s compatibility_flags [] =
-  {
-    { COMPAT_ALLOW_KA_TO_ENCR, "allow-ka-to-encr" },
-    { COMPAT_NO_CHAIN_CACHE, "no-chain-cache"     },
-    { COMPAT_NO_KEYINFO_CACHE, "no-keyinfo-cache" },
+    { DBG_CLOCK_VALUE  , "clock"   },
+    { DBG_LOOKUP_VALUE , "lookup"  },
     { 0, NULL }
   };
 
@@ -501,11 +501,8 @@ static int default_include_certs = DEFAULT_INCLUDE_CERTS;
 /* Whether the chain mode shall be used for validation.  */
 static int default_validation_model;
 
-/* Counter used to convey data from deinit_ctrl to gpgsm_exit.  */
-static unsigned int parent_cache_stats;
-
 /* The default cipher algo.  */
-#define DEFAULT_CIPHER_ALGO "AES256"
+#define DEFAULT_CIPHER_ALGO "AES"
 
 
 static char *build_list (const char *text,
@@ -576,6 +573,10 @@ our_md_test_algo (int algo)
 }
 
 
+/* nPth wrapper function definitions. */
+ASSUAN_SYSTEM_NPTH_IMPL;
+
+
 static char *
 make_libversion (const char *libname, const char *(*getfnc)(const char*))
 {
@@ -779,7 +780,7 @@ set_debug (void)
 
   /* minip12.c may be used outside of GnuPG, thus we don't have the
    * opt structure over there.  */
-  p12_set_verbosity (opt.verbose, opt.debug);
+  p12_set_verbosity (opt.verbose);
 }
 
 
@@ -844,7 +845,8 @@ parse_validation_model (const char *model)
 int
 main ( int argc, char **argv)
 {
-  ARGPARSE_ARGS pargs;
+  gpg_error_t err = 0;
+  gpgrt_argparse_t pargs;
   int orig_argc;
   char **orig_argv;
   /*  char *username;*/
@@ -882,18 +884,20 @@ main ( int argc, char **argv)
   int pwfd = -1;
 
   static const char *homedirvalue;
+  static const char *changeuser;
+
 
   early_system_init ();
   gnupg_reopen_std (GPGSM_NAME);
   /* trap_unaligned ();*/
   gnupg_rl_initialize ();
-  set_strusage (my_strusage);
+  gpgrt_set_strusage (my_strusage);
   gcry_control (GCRYCTL_SUSPEND_SECMEM_WARN);
 
   /* Please note that we may running SUID(ROOT), so be very CAREFUL
      when adding any stuff between here and the call to secmem_init()
      somewhere after the option parsing */
-  log_set_prefix (GPGSM_NAME, GPGRT_LOG_WITH_PREFIX|GPGRT_LOG_NO_REGISTRY);
+  log_set_prefix (GPGSM_NAME, GPGRT_LOG_WITH_PREFIX);
 
   /* Make sure that our subsystems are ready.  */
   i18n_init ();
@@ -934,7 +938,7 @@ main ( int argc, char **argv)
   pargs.argc = &argc;
   pargs.argv = &argv;
   pargs.flags= (ARGPARSE_FLAG_KEEP | ARGPARSE_FLAG_NOVERSION);
-  while (gnupg_argparse (NULL, &pargs, opts))
+  while (gpgrt_argparse (NULL, &pargs, opts))
     {
       switch (pargs.r_opt)
         {
@@ -953,6 +957,10 @@ main ( int argc, char **argv)
           homedirvalue = pargs.r.ret_str;
           break;
 
+        case oChUid:
+          changeuser = pargs.r.ret_str;
+          break;
+
         case aCallProtectTool:
           /* Make sure that --version and --help are passed to the
            * protect-tool. */
@@ -963,13 +971,12 @@ main ( int argc, char **argv)
   /* Reset the flags.  */
   pargs.flags &= ~(ARGPARSE_FLAG_KEEP | ARGPARSE_FLAG_NOVERSION);
 
-
   /* Initialize the secure memory. */
   gcry_control (GCRYCTL_INIT_SECMEM, 16384, 0);
   maybe_setuid = 0;
 
   /*
-   * Now we are now working under our real uid
+     Now we are now working under our real uid
    */
 
   ksba_set_malloc_hooks (gcry_malloc, gcry_realloc, gcry_free );
@@ -981,7 +988,9 @@ main ( int argc, char **argv)
   assuan_set_gpg_err_source (GPG_ERR_SOURCE_DEFAULT);
   setup_libassuan_logging (&opt.debug, NULL);
 
-  /* Set homedir.  */
+  /* Change UID and then set homedir.  */
+  if (changeuser && gnupg_chuid (changeuser, 0))
+    log_inc_errorcount (); /* Force later termination.  */
   gnupg_set_homedir (homedirvalue);
 
   /* Setup a default control structure for command line mode */
@@ -995,8 +1004,8 @@ main ( int argc, char **argv)
   opt.policy_file = make_filename (gnupg_homedir (), "policies.txt", NULL);
 
   /* The configuraton directories for use by gpgrt_argparser.  */
-  gnupg_set_confdir (GNUPG_CONFDIR_SYS, gnupg_sysconfdir ());
-  gnupg_set_confdir (GNUPG_CONFDIR_USER, gnupg_homedir ());
+  gpgrt_set_confdir (GPGRT_CONFDIR_SYS, gnupg_sysconfdir ());
+  gpgrt_set_confdir (GPGRT_CONFDIR_USER, gnupg_homedir ());
 
   /* We are re-using the struct, thus the reset flag.  We OR the
    * flags so that the internal intialized flag won't be cleared. */
@@ -1010,7 +1019,7 @@ main ( int argc, char **argv)
                    | ARGPARSE_FLAG_USER);
 
   while (!no_more_options
-         && gnupg_argparser (&pargs, opts, GPGSM_NAME EXTSEP_S "conf"))
+         && gpgrt_argparser (&pargs, opts, GPGSM_NAME EXTSEP_S "conf"))
     {
       switch (pargs.r_opt)
         {
@@ -1078,6 +1087,7 @@ main ( int argc, char **argv)
         case aExportSecretKeyP12:
         case aExportSecretKeyP8:
         case aExportSecretKeyRaw:
+        case aShowCerts:
         case aDumpKeys:
         case aDumpChain:
         case aDumpExternalKeys:
@@ -1156,10 +1166,6 @@ main ( int argc, char **argv)
           ctrl.is_base64 = 0;
           break;
 
-        case oInputSizeHint:
-          ctrl.input_size_hint = string_to_u64 (pargs.r.ret_str);
-          break;
-
         case oDisableCRLChecks:
           opt.no_crl_check = 1;
           break;
@@ -1241,6 +1247,7 @@ main ( int argc, char **argv)
         case oAnswerNo: opt.answer_no = 1; break;
 
         case oKeyring: append_to_strlist (&nrings, pargs.r.ret_str); break;
+        case oUseKeyboxd: opt.use_keyboxd = 1; break;
 
         case oDebug:
           if (parse_debug_flag (pargs.r.ret_str, &debug_value, debug_flags))
@@ -1258,15 +1265,7 @@ main ( int argc, char **argv)
           break;
         case oDebugNoChainValidation: opt.no_chain_validation = 1; break;
         case oDebugIgnoreExpiration: opt.ignore_expiration = 1; break;
-
-        case oCompatibilityFlags:
-          if (parse_compatibility_flags (pargs.r.ret_str, &opt.compat_flags,
-                                         compatibility_flags))
-            {
-              pargs.r_opt = ARGPARSE_INVALID_ARG;
-              pargs.err = ARGPARSE_PRINT_ERROR;
-            }
-          break;
+        case oDebugForceECDHSHA1KDF: opt.force_ecdh_sha1kdf = 1; break;
 
         case oStatusFD:
             ctrl.status_fd = translate_sys2libc_fd_int (pargs.r.ret_int, 1);
@@ -1286,8 +1285,14 @@ main ( int argc, char **argv)
           opt.with_keygrip = 1;
           break;
 
+        case oWithKeyScreening:
+          opt.with_key_screening = 1;
+          break;
+
         case oHomedir: gnupg_set_homedir (pargs.r.ret_str); break;
+        case oChUid: break;  /* Command line only (see above).  */
         case oAgentProgram: opt.agent_program = pargs.r.ret_str;  break;
+        case oKeyboxdProgram: opt.keyboxd_program = pargs.r.ret_str;  break;
 
         case oDisplay:
           set_opt_session_env ("DISPLAY", pargs.r.ret_str);
@@ -1413,16 +1418,14 @@ main ( int argc, char **argv)
           append_to_strlist (&opt.keyserver, pargs.r.ret_str);
 	  break;
 
-        case oKeyServer_deprecated:
-          obsolete_option (configname, pargs.lineno, "ldapserver");
-          break;
-
         case oIgnoreCertExtension:
           add_to_strlist (&opt.ignored_cert_extensions, pargs.r.ret_str);
           break;
 
-        case oIgnoreCertWithOID:
-          add_to_strlist (&opt.ignore_cert_with_oid, pargs.r.ret_str);
+        case oAuthenticode: opt.authenticode = 1; break;
+
+        case oAttribute:
+          add_to_strlist (&opt.attributes, pargs.r.ret_str);
           break;
 
         case oNoAutostart: opt.autostart = 0; break;
@@ -1434,25 +1437,15 @@ main ( int argc, char **argv)
                 { "gnupg", CO_GNUPG },
                 { "de-vs", CO_DE_VS }
               };
-            int compliance = gnupg_parse_compliance_option (pargs.r.ret_str,
-                                                            compliance_options,
-                                                            DIM (compliance_options),
-                                                            opt.quiet);
+            int compliance = gnupg_parse_compliance_option
+              (pargs.r.ret_str, compliance_options, DIM (compliance_options),
+               opt.quiet);
             if (compliance < 0)
               log_inc_errorcount (); /* Force later termination.  */
             opt.compliance = compliance;
           }
           break;
 
-        case oMinRSALength: opt.min_rsa_length = pargs.r.ret_ulong; break;
-
-        case oRequireCompliance: opt.require_compliance = 1;  break;
-        case oAlwaysTrust: opt.always_trust = 1;  break;
-
-        case oKbxBufferSize:
-          keybox_set_buffersize (pargs.r.ret_ulong, 0);
-          break;
-
         default:
           if (configname)
             pargs.err = ARGPARSE_PRINT_WARNING;
@@ -1468,12 +1461,12 @@ main ( int argc, char **argv)
 	}
     }
 
-  gnupg_argparse (NULL, &pargs, NULL);  /* Release internal state.  */
+  gpgrt_argparse (NULL, &pargs, NULL);  /* Release internal state.  */
 
   if (!last_configname)
-    opt.config_filename = make_filename (gnupg_homedir (),
-                                         GPGSM_NAME EXTSEP_S "conf",
-                                         NULL);
+    opt.config_filename = gpgrt_fnameconcat (gnupg_homedir (),
+                                             GPGSM_NAME EXTSEP_S "conf",
+                                             NULL);
   else
     opt.config_filename = last_configname;
 
@@ -1484,6 +1477,28 @@ main ( int argc, char **argv)
       gpgsm_exit(2);
     }
 
+  /* Process common component options.  */
+  if (parse_comopt (GNUPG_MODULE_NAME_GPGSM, debug_argparser))
+    {
+      gpgsm_status_with_error (&ctrl, STATUS_FAILURE,
+                               "option-parser", gpg_error (GPG_ERR_GENERAL));
+      gpgsm_exit(2);
+    }
+
+  if (opt.use_keyboxd)
+    log_info ("Note: Please move option \"%s\" to \"common.conf\"\n",
+              "use-keyboxd");
+  opt.use_keyboxd = comopt.use_keyboxd;  /* Override.  */
+
+  if (opt.keyboxd_program)
+    log_info ("Note: Please move option \"%s\" to \"common.conf\"\n",
+              "keyboxd-program");
+  if (!opt.keyboxd_program && comopt.keyboxd_program)
+    {
+      opt.keyboxd_program = comopt.keyboxd_program;
+      comopt.keyboxd_program = NULL;
+    }
+
   if (pwfd != -1)	/* Read the passphrase now.  */
     read_passphrase_from_fd (pwfd);
 
@@ -1497,29 +1512,24 @@ main ( int argc, char **argv)
   if (greeting)
     {
       es_fprintf (es_stderr, "%s %s; %s\n",
-                  strusage(11), strusage(13), strusage(14) );
-      es_fprintf (es_stderr, "%s\n", strusage(15) );
+                  gpgrt_strusage(11), gpgrt_strusage(13), gpgrt_strusage(14) );
+      es_fprintf (es_stderr, "%s\n", gpgrt_strusage(15) );
     }
-#  ifdef IS_DEVELOPMENT_VERSION
+#ifdef IS_DEVELOPMENT_VERSION
   if (!opt.batch)
     {
       log_info ("NOTE: THIS IS A DEVELOPMENT VERSION!\n");
       log_info ("It is only intended for test purposes and should NOT be\n");
       log_info ("used in a production environment or with production keys!\n");
     }
-#  endif
+#endif
 
   if (may_coredump && !opt.quiet)
     log_info (_("WARNING: program may create a core file!\n"));
 
-  if (opt.require_compliance && opt.always_trust)
-    {
-      opt.always_trust = 0;
-      if (opt.quiet)
-        log_info (_("WARNING: %s overrides %s\n"),
-                  "--require-compliance","--always-trust");
-    }
-
+  npth_init ();
+  assuan_set_system_hooks (ASSUAN_SYSTEM_NPTH);
+  gpgrt_set_syscall_clamp (npth_unprotect, npth_protect);
 
 /*   if (opt.qualsig_approval && !opt.quiet) */
 /*     log_info (_("This software has officially been approved to " */
@@ -1558,11 +1568,8 @@ main ( int argc, char **argv)
   gcry_control (GCRYCTL_RESUME_SECMEM_WARN);
 
   set_debug ();
-  if (opt.verbose) /* Print the compatibility flags.  */
-    parse_compatibility_flags (NULL, &opt.compat_flags, compatibility_flags);
-  gnupg_set_compliance_extra_info (CO_EXTRA_INFO_MIN_RSA, opt.min_rsa_length);
 
-  /* Although we always use gpgsm_exit, we better install a regualr
+  /* Although we always use gpgsm_exit, we better install a regular
      exit handler so that at least the secure memory gets wiped
      out. */
   if (atexit (emergency_cleanup))
@@ -1678,8 +1685,16 @@ main ( int argc, char **argv)
   if (!cmd && opt.fingerprint && !with_fpr)
     set_cmd (&cmd, aListKeys);
 
+  /* If no pinentry is expected shunt
+   * gnupg_allow_set_foregound_window to avoid useless error
+   * messages on Windows.  */
+  if (opt.pinentry_mode != PINENTRY_MODE_ASK)
+    {
+      gnupg_inhibit_set_foregound_window (1);
+    }
+
   /* Add default keybox. */
-  if (!nrings && default_keyring)
+  if (!nrings && default_keyring && !opt.use_keyboxd)
     {
       int created;
 
@@ -1700,8 +1715,11 @@ main ( int argc, char **argv)
           xfree (filelist[0]);
         }
     }
-  for (sl = nrings; sl; sl = sl->next)
-    keydb_add_resource (&ctrl, sl->d, 0, NULL);
+  if (!opt.use_keyboxd)
+    {
+      for (sl = nrings; sl; sl = sl->next)
+        keydb_add_resource (&ctrl, sl->d, 0, NULL);
+    }
   FREE_STRLIST(nrings);
 
 
@@ -1777,7 +1795,7 @@ main ( int argc, char **argv)
   switch (cmd)
     {
     case aGPGConfList:
-      { /* List options and default values in the GPG Conf format.  */
+      { /* List default option values in the GPG Conf format.  */
 
 	es_printf ("debug-level:%lu:\"none:\n", GC_OPT_FLAG_DEFAULT);
         es_printf ("include-certs:%lu:%d:\n", GC_OPT_FLAG_DEFAULT,
@@ -1792,6 +1810,7 @@ main ( int argc, char **argv)
            proc_parameters actually implements.  */
         es_printf ("default_pubkey_algo:%lu:\"%s:\n", GC_OPT_FLAG_DEFAULT,
                    "RSA-3072");
+
       }
       break;
     case aGPGConfTest:
@@ -1829,13 +1848,16 @@ main ( int argc, char **argv)
         set_binary (stdin);
 
         if (!argc) /* Source is stdin. */
-          gpgsm_encrypt (&ctrl, recplist, 0, fp);
+          err = gpgsm_encrypt (&ctrl, recplist, 0, fp);
         else if (argc == 1)  /* Source is the given file. */
-          gpgsm_encrypt (&ctrl, recplist, open_read (*argv), fp);
+          err = gpgsm_encrypt (&ctrl, recplist, open_read (*argv), fp);
         else
           wrong_args ("--encrypt [datafile]");
 
-        es_fclose (fp);
+        if (err)
+          gpgrt_fcancel (fp);
+        else
+          es_fclose (fp);
       }
       break;
 
@@ -1847,29 +1869,31 @@ main ( int argc, char **argv)
            signing because that is what gpg does.*/
         set_binary (stdin);
         if (!argc) /* Create from stdin. */
-          gpgsm_sign (&ctrl, signerlist, 0, detached_sig, fp);
+          err = gpgsm_sign (&ctrl, signerlist, 0, detached_sig, fp);
         else if (argc == 1) /* From file. */
-          gpgsm_sign (&ctrl, signerlist,
+          err = gpgsm_sign (&ctrl, signerlist,
                       open_read (*argv), detached_sig, fp);
         else
           wrong_args ("--sign [datafile]");
 
+#if GPGRT_VERSION_NUMBER >= 0x012700 /* >= 1.39 */
+        if (err)
+          gpgrt_fcancel (fp);
+        else
+          es_fclose (fp);
+#else
+        (void)err;
         es_fclose (fp);
+#endif
       }
       break;
 
     case aSignEncr: /* sign and encrypt the given file */
-      log_error ("the command '%s' has not yet been implemented\n",
-                 "--sign --encrypt");
-      gpgsm_status_with_error (&ctrl, STATUS_FAILURE, "option-parser",
-                               gpg_error (GPG_ERR_NOT_IMPLEMENTED));
+      log_error ("this command has not yet been implemented\n");
       break;
 
     case aClearsign: /* make a clearsig */
-      log_error ("the command '%s' has not yet been implemented\n",
-                 "--clearsign");
-      gpgsm_status_with_error (&ctrl, STATUS_FAILURE, "option-parser",
-                               gpg_error (GPG_ERR_NOT_IMPLEMENTED));
+      log_error ("this command has not yet been implemented\n");
       break;
 
     case aVerify:
@@ -1898,7 +1922,6 @@ main ( int argc, char **argv)
     case aDecrypt:
       {
         estream_t fp = open_es_fwrite (opt.outfile?opt.outfile:"-");
-        gpg_error_t err;
 
         set_binary (stdin);
         if (!argc)
@@ -1908,11 +1931,9 @@ main ( int argc, char **argv)
         else
           wrong_args ("--decrypt [filename]");
 
-#if GPGRT_VERSION_NUMBER >= 0x012700  /* 1.39 */
         if (err)
           gpgrt_fcancel (fp);
         else
-#endif
           es_fclose (fp);
       }
       break;
@@ -1959,6 +1980,15 @@ main ( int argc, char **argv)
       }
       break;
 
+    case aShowCerts:
+      {
+        estream_t fp;
+
+        fp = open_es_fwrite (opt.outfile?opt.outfile:"-");
+        gpgsm_show_certs (&ctrl, argc, argv, fp);
+        es_fclose (fp);
+      }
+      break;
 
     case aKeygen: /* Generate a key; well kind of. */
       {
@@ -2096,10 +2126,8 @@ main ( int argc, char **argv)
 
 
     default:
-      log_error (_("invalid command (there is no implicit command)\n"));
-      gpgsm_status_with_error (&ctrl, STATUS_FAILURE, "option-parser",
-                               gpg_error (GPG_ERR_MISSING_ACTION));
-      break;
+        log_error (_("invalid command (there is no implicit command)\n"));
+	break;
     }
 
   /* Print the audit result if needed.  */
@@ -2141,7 +2169,6 @@ gpgsm_exit (int rc)
   gcry_control (GCRYCTL_UPDATE_RANDOM_SEED_FILE);
   if (opt.debug & DBG_MEMSTAT_VALUE)
     {
-      log_info ("cert_chain_cache: cached=%u\n", parent_cache_stats);
       gcry_control( GCRYCTL_DUMP_MEMORY_STATS );
       gcry_control( GCRYCTL_DUMP_RANDOM_STATS );
     }
@@ -2168,20 +2195,7 @@ gpgsm_init_default_ctrl (struct server_control_s *ctrl)
 void
 gpgsm_deinit_default_ctrl (ctrl_t ctrl)
 {
-  unsigned int n;
-
-  gpgsm_flush_keyinfo_cache (ctrl);
-  n = 0;
-  while (ctrl->parent_cert_cache)
-    {
-      cert_cache_item_t next = ctrl->parent_cert_cache->next;
-      ksba_cert_release (ctrl->parent_cert_cache->result);
-      xfree (ctrl->parent_cert_cache);
-      ctrl->parent_cert_cache = next;
-      n++;
-    }
-  if (n > parent_cache_stats)
-    parent_cache_stats = n;
+  gpgsm_keydb_deinit_session_data (ctrl);
 }
 
 
